@@ -200,16 +200,9 @@ _gpgme_wait_global_event_cb (void *data, gpgme_event_io_t type,
 	gpgme_error_t err = ctx_active (ctx);
 
 	if (err)
-	  {
-	    /* An error occured.  Close all fds in this context, and
-	       send the error in a done event.  */
-	    unsigned int idx;
-	    
-	    for (idx = 0; idx <= ctx->fdt.size; idx++)
-	      if (ctx->fdt.fds[idx].fd != -1)
-		_gpgme_io_close (ctx->fdt.fds[idx].fd);
-	    _gpgme_engine_io_event (ctx->engine, GPGME_EVENT_DONE, &err);
-	  }
+	  /* An error occured.  Close all fds in this context, and
+	     send the error in a done event.  */
+	  _gpgme_cancel_with_err (ctx, &err);
       }
       break;
 
@@ -299,7 +292,7 @@ gpgme_wait (gpgme_ctx_t ctx, gpgme_error_t *status, int hang)
 	  if (fdt.fds[i].fd != -1 && fdt.fds[i].signaled)
 	    {
 	      gpgme_ctx_t ictx;
-	      gpgme_error_t err;
+	      gpgme_error_t err = 0;
 	      struct wait_item_s *item;
 	      
 	      assert (nr);
@@ -310,18 +303,18 @@ gpgme_wait (gpgme_ctx_t ctx, gpgme_error_t *status, int hang)
 	      ictx = item->ctx;
 	      assert (ictx);
 
-	      err = _gpgme_run_io_cb (&fdt.fds[i], 0);
+	      LOCK (ctx->lock);
+	      if (ctx->canceled)
+		err = gpg_error (GPG_ERR_CANCELED);
+	      UNLOCK (ctx->lock);
+
+	      if (!err)
+		err = _gpgme_run_io_cb (&fdt.fds[i], 0);
 	      if (err)
 		{
 		  /* An error occured.  Close all fds in this context,
 		     and signal it.  */
-		  unsigned int idx;
-	    
-		  for (idx = 0; idx < ictx->fdt.size; idx++)
-		    if (ictx->fdt.fds[idx].fd != -1)
-		      _gpgme_io_close (ictx->fdt.fds[idx].fd);
-		  _gpgme_engine_io_event (ictx->engine, GPGME_EVENT_DONE,
-					  &err);
+		  _gpgme_cancel_with_err (ictx, err);
 
 		  /* Break out of the loop, and retry the select()
 		     from scratch, because now all fds should be
@@ -334,6 +327,7 @@ gpgme_wait (gpgme_ctx_t ctx, gpgme_error_t *status, int hang)
 
       /* Now some contexts might have finished successfully.  */
       LOCK (ctx_list_lock);
+    retry:
       for (li = ctx_active_list; li; li = li->next)
 	{
 	  gpgme_ctx_t actx = li->ctx;
@@ -344,7 +338,19 @@ gpgme_wait (gpgme_ctx_t ctx, gpgme_error_t *status, int hang)
 	  if (i == actx->fdt.size)
 	    {
 	      gpgme_error_t err = 0;
+
+	      /* FIXME: This does not perform too well.  We have to
+		 release the lock because the I/O event handler
+		 acquires it to remove the context from the active
+		 list.  Two alternative strategies are worth
+		 considering: Either implement the DONE event handler
+		 here in a lock-free manner, or save a list of all
+		 contexts to be released and call the DONE events
+		 afterwards.  */
+	      UNLOCK (ctx_list_lock);
 	      _gpgme_engine_io_event (actx->engine, GPGME_EVENT_DONE, &err);
+	      LOCK (ctx_list_lock);
+	      goto retry;
 	    }
 	}
       UNLOCK (ctx_list_lock);
